from .competition import Competition

from .groups import RoundRobin, SwissRound
from .random import RandomRound
from .swissbracket import SwissBracket

from .knockout import SingleEliminationBracket, DoubleEliminationBracket
from .snake import Snake
